# read version from installed package
from importlib.metadata import version
from bias_adjustment.bias_adjustment import BiasAdjustment

__version__ = version("bias_adjustment")
__all__ = [BiasAdjustment]



import warnings
from dataclasses import dataclass
from typing import List

import numpy as np
from scipy import stats as st

from bias_adjustment.utils import FloatNDArray, is_float_ndarray


def _fit_hist(data: FloatNDArray, bins=200):
    """Generate a distribution from a histogram

    Args:
        data (FloatNDArray): Input data
        bins (int, optional): Number of bins. Defaults to 200.
    """

    for v in [data]:
        if not is_float_ndarray(v):
            raise TypeError(f"`{v}` is not a float numpy array.")
        min_len = 10
        if len(v) < min_len:
            raise ValueError(f"Length of `{v}` must be greater than {min_len}.")

    for v in [bins]:
        if not isinstance(v, int):
            raise TypeError(f"`{v}` is not an integer.")

    h = np.histogram(data[~np.isnan(data)], bins=bins)
    return st.rv_histogram(h)


def _fit_dist(data: FloatNDArray, dist_type="gamma"):
    """Generate a distribution given a distribution type

    Args:
        data (FloatNDArray): Input data.
        dist_type (str, optional): Valid scipy.stats distribution name. Defaults to "gamma".
    """

    for v in [data]:
        if not is_float_ndarray(v):
            raise TypeError(f"`{v}` is not a float numpy array.")
        min_len = 10
        if len(v) < min_len:
            raise ValueError(f"Length of `{v}` must be greater than {min_len}.")

    if not (
        hasattr(st, dist_type) and isinstance(getattr(st, dist_type), st.rv_continuous)
    ):
        raise ValueError("`dist_type` must be a valid scipy.stats distribution name.")

    dist: st.rv_continuous = getattr(st, dist_type)
    params = dist.fit(data[~np.isnan(data)])
    return dist(*params)


@dataclass
class Distributions:
    data: FloatNDArray
    min_len: int = 10

    def __post_init__(self):
        for name in ["data"]:
            attr = getattr(self, name)
            if not is_float_ndarray(attr):
                raise TypeError(f"`{name}` is not a float numpy array.")
            if len(attr) < self.min_len:
                raise ValueError(
                    f"Length of `{name}` must be greater than {self.min_len}."
                )

        for name in ["min_len"]:
            attr = getattr(self, name)
            if not isinstance(attr, int):
                raise TypeError(f"`{name}` is not an integer.")

    def fit(self, dist_type="hist", bins=200):
        """Generate distribution

        Args:
            dist_type (str, optional): Valid scipy.stats distribution name. Defaults to "hist".
            bins (int, optional): Number of bins. Defaults to 200.
        """

        if not isinstance(dist_type, str):
            raise TypeError(
                "`dist_type` must be a valid scipy.stats distribution name."
            )

        if not isinstance(bins, int):
            raise TypeError("`bins` must be an integer.")

        if dist_type == "hist":
            return _fit_hist(self.data, bins)
        elif hasattr(st, dist_type) and isinstance(
            getattr(st, dist_type), st.rv_continuous
        ):
            return _fit_dist(self.data, dist_type)
        else:
            raise ValueError(
                "`dist_type` must be a valid scipy.stats distribution name."
            )

    def best_fit(self, distributions: List[str] = None, bins=100):
        """Find best fit distribution to data

        Args:
            distributions (List[str], optional): List of distribution names to test.
                Defaults to ["norm", "lognorm", "gamma"].
            bins (int, optional): Number of bins. Defaults to 100.
        """

        if distributions is None:
            distributions = ["norm", "lognorm", "gamma"]
        elif not (
            isinstance(distributions, list)
            and all(isinstance(elem, str) for elem in distributions)
        ):
            raise TypeError(
                "`distributions` must be a valid scipy.stats distribution name."
            )

        if not isinstance(bins, int):
            raise TypeError("`bins` must be an integer.")

        # Get histogram of original data
        _data = self.data[~np.isnan(self.data)]
        y, x = np.histogram(_data, bins=bins, density=True)
        x = (x + np.roll(x, -1))[:-1] / 2.0

        # Best holders
        best_dist: st.rv_continuous = getattr(st, "norm")
        best_sse = np.inf

        for dist_name in distributions:
            # Try to fit the distribution
            # print(dist_name)
            try:
                # fit dist to data
                dist = self.fit(dist_type=dist_name, bins=bins)

                # Calculate fitted PDF and error with fit in distribution
                pdf = dist.pdf(x)
                sse = np.sum(np.power(y - pdf, 2.0))

                # identify if this distribution is better
                if best_sse > sse > 0:
                    best_dist = dist
                    best_sse = sse
            except Exception as e:
                warnings.warn(str(e))
                continue
            # print(dist, params)

        return best_dist



from typing import Literal, Union, get_args

import numpy as np
import numpy.typing as npt

FloatNDArray = npt.NDArray[np.float64]

ARRAY_LIKE = Union[list, np.ndarray]

BAMode = Literal["rel", "abs"]


def is_float_ndarray(v):
    return isinstance(v, np.ndarray) and np.issubdtype(v.dtype, np.floating)


def is_array_like(v):
    return isinstance(v, get_args(ARRAY_LIKE))



from dataclasses import dataclass

import numpy as np

from bias_adjustment.distributions import Distributions
from bias_adjustment.utils import FloatNDArray, is_float_ndarray


@dataclass
class QuantileMapping:
    obs: FloatNDArray
    mod: FloatNDArray
    data: FloatNDArray
    max_cdf: float = 0.99999

    def __post_init__(self):
        for name in ["obs", "mod", "data"]:
            attr = getattr(self, name)
            if not is_float_ndarray(attr):
                raise TypeError(f"`{name}` is not a numpy array.")
            min_len = 10
            if len(attr) < min_len:
                raise ValueError(f"Length of `{name}` must be greater than {min_len}.")

        for name in ["max_cdf"]:
            attr = getattr(self, name)
            if not isinstance(attr, float):
                raise TypeError(f"`{name}` is not a float.")
            if name == "max_cdf":
                if attr < 0.5 or attr >= 1:
                    raise ValueError((f"`{name}` should be [0.5, 1)."))

    @staticmethod
    def generate_distribution(data: FloatNDArray, dist_type="hist"):
        """Generate Distribution
        Args:
            data (FloatNDArray): Input data.
            dist_type (str, optional): Valid scipy.stats distribution name. Defaults to "hist".
        """
        return Distributions(data).fit(dist_type)

    def compute(
        self,
        dist_type="hist",
    ) -> FloatNDArray:
        """Adjust the bias

        Args:
            dist_type (str, optional): Valid scipy.stats distribution name. Defaults to "hist".

        Returns:
            FloatNDArray: The adjusted values.
        """
        o_dist = self.generate_distribution(self.obs, dist_type)
        m_dist = self.generate_distribution(self.mod, dist_type)

        m_cdf = np.minimum(self.max_cdf, m_dist.cdf(self.data))
        return o_dist.ppf(m_cdf)



from bias_adjustment.quantile_mapping.qm import QuantileMapping
from bias_adjustment.quantile_mapping.dqm import DetrendedQuantileMapping
from bias_adjustment.quantile_mapping.qdm import QuantileDeltaMapping

__all__ = [
    QuantileMapping,
    DetrendedQuantileMapping,
    QuantileDeltaMapping,
]



from dataclasses import dataclass
from typing import get_args

import numpy as np

from bias_adjustment.quantile_mapping.qm import QuantileMapping
from bias_adjustment.utils import BAMode, FloatNDArray


@dataclass
class QuantileDeltaMapping(QuantileMapping):
    def compute(
        self,
        mode: BAMode = "rel",
        dist_type="hist",
    ) -> FloatNDArray:
        if mode not in get_args(BAMode):
            raise ValueError(f"Length of `mode` must be {get_args(BAMode)}.")

        o_dist = self.generate_distribution(self.obs, dist_type)
        mh_dist = self.generate_distribution(self.mod, dist_type)
        mf_dist = self.generate_distribution(self.data, dist_type)

        mf_cdf = np.minimum(self.max_cdf, mf_dist.cdf(self.data))

        if mode == "rel":  # Relative
            return o_dist.ppf(mf_cdf) * (self.data / mh_dist.ppf(mf_cdf))
        elif mode == "abs":  # Absolute
            return o_dist.ppf(mf_cdf) + self.data - mh_dist.ppf(mf_cdf)
        else:
            return None



from dataclasses import dataclass
from typing import get_args

import numpy as np

from bias_adjustment.quantile_mapping.qm import QuantileMapping
from bias_adjustment.utils import BAMode, FloatNDArray


@dataclass
class DetrendedQuantileMapping(QuantileMapping):
    def delta(self, mode: BAMode = "rel"):
        mod_mean = np.nanmean(self.mod)
        dat_mean = np.nanmean(self.data)
        if mode == "rel":
            return dat_mean / mod_mean
        elif mode == "abs":
            return dat_mean - mod_mean
        return None

    def compute(
        self,
        mode: BAMode = "rel",
        dist_type="hist",
    ) -> FloatNDArray:
        if mode not in get_args(BAMode):
            raise ValueError(f"Length of `mode` must be {get_args(BAMode)}.")

        o_dist = self.generate_distribution(self.obs, dist_type)
        m_dist = self.generate_distribution(self.mod, dist_type)

        delta = self.delta(mode)
        if mode == "rel":
            m_cdf = np.minimum(self.max_cdf, m_dist.cdf(self.data / delta))
            return o_dist.ppf(m_cdf) * delta
        elif mode == "abs":
            m_cdf = np.minimum(self.max_cdf, m_dist.cdf(self.data - delta))
            return o_dist.ppf(m_cdf) + delta
        return o_dist.ppf(m_cdf)



from dataclasses import dataclass

from bias_adjustment.quantile_mapping import (
    DetrendedQuantileMapping,
    QuantileDeltaMapping,
    QuantileMapping,
)
from bias_adjustment.utils import FloatNDArray, is_float_ndarray


def _get_ba_mode(method: str):
    args = method.split(".")
    if len(args) == 2:
        return args[1]
    return


@dataclass
class BiasAdjustment:
    obs: FloatNDArray
    mod: FloatNDArray
    max_cdf: float = 0.99999

    def __post_init__(self):
        for name in ["obs", "mod"]:
            attr = getattr(self, name)
            if not is_float_ndarray(attr):
                raise TypeError(f"`{name}` is not a numpy array.")
            min_len = 10
            if len(attr) < min_len:
                raise ValueError(f"Length of `{name}` must be greater than {min_len}.")

        for name in ["max_cdf"]:
            attr = getattr(self, name)
            if not isinstance(attr, float):
                raise TypeError(f"`{name}` is not a float.")
            if name == "max_cdf":
                if attr < 0.5 or attr >= 1:
                    raise ValueError((f"`{name}` should be [0.5, 1)."))

    def adjust(
        self,
        data: FloatNDArray,
        method="qm",
        dist_type="hist",
    ):
        if method == "qm":
            return QuantileMapping(
                self.obs, self.mod, data, max_cdf=self.max_cdf
            ).compute(dist_type=dist_type)
        elif method.startswith("dqm"):
            mode = _get_ba_mode(method)
            qm = DetrendedQuantileMapping(
                self.obs, self.mod, data, max_cdf=self.max_cdf
            )
            if mode is not None:
                return qm.compute(
                    mode=mode,
                    dist_type=dist_type,
                )
            return qm.compute(
                dist_type=dist_type,
            )
        elif method.startswith("qdm"):
            mode = _get_ba_mode(method)
            qm = QuantileDeltaMapping(self.obs, self.mod, data, max_cdf=self.max_cdf)
            if mode is not None:
                return qm.compute(
                    mode=mode,
                    dist_type=dist_type,
                )
            return qm.compute(
                dist_type=dist_type,
            )
        return QuantileMapping(self.obs, self.mod, data, max_cdf=self.max_cdf).compute(
            dist_type=dist_type
        )
